from . import _read_uint32_le
from .ReiaFrame import ReiaFrame, create_frame_reader
import typing


class ReiaFile:
    """A .reia video file.
    Note that all frames are held uncompressed in memory.

    Attributes
    ------------

    width
        The horizontal resolution of the video in pixels.

    height
        The vertical resolution of the video in pixels.

    frames_per_second
        The framerate of the video.

    frames
        An iterator over the individual frames of the video. Note that this
        can only be iterated over once, so save any frames you need or turn
        this into a list.
    """

    width: int
    height: int
    frames_per_second: float
    num_frames: int
    frames: typing.Iterator[ReiaFrame]

    def __init__(self, width, height, frames_per_second, num_frames, frames) -> None:
        # Assert that width and height are divisible by 32.
        assert (width % 32) == 0
        self.width = width
        assert (height % 32) == 0
        self.height = height

        self.frames_per_second = frames_per_second
        self.num_frames = num_frames
        self.frames = frames


def read_from_file(stream: typing.BinaryIO) -> ReiaFile:
    # Assert that the file starts with the proper magic bytes.
    riff_file_magic = stream.read(4)
    if riff_file_magic != b"RIFF":
        raise ValueError(
            f"Incorrect magic at start of file, expected 'RIFF', got {riff_file_magic}"
        )

    # We can check that matches up, but unused for now.
    file_size = _read_uint32_le(stream)

    reia_header_magic = stream.read(8)
    if reia_header_magic != b"Reiahead":
        raise ValueError(
            f"Incorrect magic inside RIFF container, expected 'Reiahead', got {reia_header_magic}"
        )

    size_of_metadata = _read_uint32_le(stream)
    if size_of_metadata != 24:
        raise ValueError(f"Reiahead metadata size not 24, got {size_of_metadata}")

    # This value is checked to be always 1 in the real code. Just assert here.
    assert _read_uint32_le(stream) == 1

    width = _read_uint32_le(stream)
    height = _read_uint32_le(stream)

    # Frames per second.
    frames_per_second_numerator = _read_uint32_le(stream)
    frames_per_second_denominator = _read_uint32_le(stream)
    frames_per_second = (
        float(frames_per_second_numerator) / frames_per_second_denominator
    )

    # Read the expected number of frames.
    num_frames = _read_uint32_le(stream)
    frames = create_frame_reader(stream, width, height)

    return ReiaFile(width, height, frames_per_second, num_frames, frames)



from . import _read_uint32_le
from PIL import Image, ImageChops
import typing


class ReiaFrame:
    """A single frame of video."""

    image: Image

    def __init__(self, image) -> None:
        self.image = image


def read_single_pixel(stream: typing.BinaryIO) -> bytes:
    # We reverse here with `::-1` because the RGB value is stored as little endian.
    pixel_value = stream.read(3)[::-1]
    return pixel_value


def read_32_by_32_pixel_block(stream: typing.BinaryIO) -> Image:
    num_pixels = 32 * 32
    # 3 bytes for the RGB channels per pixel.
    image_data = bytearray(num_pixels * 3)

    i = 0
    while i < num_pixels:
        rle_byte = stream.read(1)
        assert rle_byte != b""

        rle_value = int.from_bytes(rle_byte, byteorder="big", signed=True)

        if rle_value < 0:
            # Negative RLE value means we are going to be repeating the next
            # color -n times.
            num_repeats = -rle_value
            pixel = read_single_pixel(stream)
            for _ in range(num_repeats + 1):
                image_data[(i * 3) : (i * 3) + 3] = pixel
                i += 1
        else:
            # Positive RLE value means we are going to be getting n unique
            # pixels.
            num_unique_pixels = rle_value
            for _ in range(num_unique_pixels + 1):
                pixel = read_single_pixel(stream)
                image_data[(i * 3) : (i * 3) + 3] = pixel
                i += 1

    return Image.frombytes("RGB", (32, 32), bytes(image_data))


def read_single_frame(
    stream: typing.BinaryIO, width: int, height: int, previous_frame: Image
) -> Image:
    image = Image.new("RGB", (width, height))

    # Frames are encoded as blocks of 32x32 pixels.
    for i in range(width // 32):
        for j in range(height // 32):
            # x and y coordinates where the top-left corner is 0,0
            x, y = (j * 32), (i * 32)
            # First bit tells us if we should expect a new 32x32 pixel block or re-use
            # the one from the previous frame.
            if stream.read(1) != b"\x00":
                block = read_32_by_32_pixel_block(stream)
                if previous_frame is not None:
                    previous_block = previous_frame.image.crop((x, y, x + 32, y + 32))
                    # Compute `(x+y) & 0xFF` for each pixel in this block and
                    # the previous block. (The blocks encode the diff from the
                    # previous frame to the current frame).
                    block = ImageChops.add_modulo(block, previous_block)
                image.paste(block, (x, y))
                continue

            # Okay, try to re-use the previous block :)
            if previous_frame is None:
                raise ValueError("32x32 block not sent but no previous frame")
            last_block = previous_frame.image.crop((x, y, x + 32, y + 32))
            image.paste(last_block, (x, y))

    return ReiaFrame(image)


def create_frame_reader(
    stream: typing.BinaryIO, width: int, height: int
) -> typing.Iterator[ReiaFrame]:
    """Returns a generator that will return Reia frames from the given stream
    at a particular width and height."""
    previous_frame = None

    # Keep reading frames until the end of the file.
    frame_magic = stream.read(4)
    while frame_magic != b"":
        # Make sure this is a valid start-of-frame.
        if frame_magic != b"frme":
            raise ValueError(
                f"Unexpected magic in start-of-frame, expected 'frme' got {frame_magic}"
            )

        frame_size = _read_uint32_le(stream)
        frame = read_single_frame(stream, width, height, previous_frame)
        previous_frame = frame
        yield frame

        # Frames get padded to be aligned on 2-byte boundaries, so consume the
        # padding (if any).
        padding = frame_size % 2
        stream.read(padding)

        # Read the next set of magic.
        frame_magic = stream.read(4)


def read_frames(
    stream: typing.BinaryIO, width: int, height: int
) -> typing.List[ReiaFrame]:
    """A convenience non-generated version of create_frame_reader that holds all
    frames in memory.
    """
    return list(create_frame_reader(stream, width, height))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sims_reia']

package_data = \
{'': ['*']}

install_requires = \
['pillow>=9.3.0,<10.0.0']

setup_kwargs = {
    'name': 'sims-reia',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Sims2 Reia File Library\n\nThis is a parser for Sims 2 `.reia` files in Python.\n\n## Basic Usage\n\nRead a `.reia` file:\n\n```python\nwith open("N001.reia", "rb") as f:\n    reia_file = sims_reia.read_from_file(input)\n    print(f"resolution={reia_file.width}x{reia_file.height}, fps={reia_file.frames_per_second}")\n\n    for i, frame in enuemrate(reia_file.frames):\n        frame.image.save(f"frame{i}.png")\n```\n\n## Testing\n\n`poetry run pytest`\n\n## Formatting\n\n`poetry run black .`\n\n',
    'author': 'Ammar Askar',
    'author_email': 'ammar@ammaraskar.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



def _read_uint32_le(stream) -> int:
    """Read a 32-bit little endian unsigned integer."""
    return int.from_bytes(stream.read(4), byteorder="little", signed=False)


from .ReiaFile import ReiaFile, read_from_file
from .ReiaFrame import ReiaFrame
from .encoder import write_reia_file



from .ReiaFile import ReiaFile
from .ReiaFrame import ReiaFrame

from PIL import Image, ImageChops

import typing


def pack_uint32_le(value: int) -> bytes:
    """Encodes an int as a 32-bit little endian unsigned integer."""
    return value.to_bytes(4, byteorder="little", signed=False)


def write_reia_file(file: ReiaFile, output_stream: typing.BinaryIO):
    assert output_stream.seekable()

    # Write the magic for the RIFF container header.
    output_stream.write(b"RIFF")

    # Length of the generated output, we write a placeholder for now and then
    # seek back here to write the real length.
    output_stream.write(pack_uint32_le(1337))

    output_stream.write(b"Reiahead")
    # Size of the metadata to follow:
    #    4 bytes for unknown field (always set to 1)
    #    4 bytes for width
    #    4 bytes for height
    #    4 bytes for fps numerator
    #    4 bytes for fps denominator
    #  + 4 bytes for number of frames
    #  ------------------------------
    #  = 24
    output_stream.write(pack_uint32_le(24))
    # Unknown field.
    output_stream.write(pack_uint32_le(1))
    # Video width.
    output_stream.write(pack_uint32_le(file.width))
    # Video height.
    output_stream.write(pack_uint32_le(file.height))

    # Calculate the fps numerator and denominator.
    #
    # For compatibility with the game files we choose the same numerator as them
    # for known frame rates.
    fps_numerator = 1_000_000
    if file.frames_per_second == 10:
        fps_numerator = 10
    fps_denominator = int(fps_numerator // file.frames_per_second)
    # Frames per second data.
    output_stream.write(pack_uint32_le(fps_numerator))
    output_stream.write(pack_uint32_le(fps_denominator))
    # Number of frames.
    output_stream.write(pack_uint32_le(file.num_frames))

    # Write out the frames.
    write_reia_frames(file.frames, output_stream)

    # Seek back to the start of the file and write the RIFF container size
    # properly.
    file_size = output_stream.tell()
    output_stream.seek(4)
    # Size of what we wrote minus the magic and the size field itself.
    output_stream.write(pack_uint32_le(file_size - 8))


def write_reia_frames(
    frames: typing.Iterator[ReiaFrame], output_stream: typing.BinaryIO
):
    previous_frame_image = None
    for frame in frames:
        encoded_frame = write_reia_frame(frame.image, previous_frame_image)
        output_stream.write(b"frme")
        output_stream.write(pack_uint32_le(len(encoded_frame)))
        output_stream.write(encoded_frame)
        # Add padding to align frames to nearest 2-byte boundary if needed.
        if len(encoded_frame) % 2 != 0:
            output_stream.write(b"\x00")

        # Make sure they're all the same resolution!
        if previous_frame_image is not None:
            assert frame.image.size == previous_frame_image.size
        previous_frame_image = frame.image


def write_reia_frame(frame, previous_frame) -> bytes:
    output = bytearray()

    width, height = frame.size[0], frame.size[1]

    assert (width % 32) == 0
    assert (height % 32) == 0

    # Frames are encoded as blocks of 32x32 pixels.
    for i in range(width // 32):
        for j in range(height // 32):
            # x and y coordinates where the top-left corner is 0,0
            x, y = (j * 32), (i * 32)

            current_block = frame.crop((x, y, x + 32, y + 32))
            previous_block = None
            if previous_frame:
                previous_block = previous_frame.crop((x, y, x + 32, y + 32))

            block = write_reia_block(current_block, previous_block)
            output.extend(block)

    return output


def find_identical_runs(raw_bytes: bytes):
    """Finds consecutive 24-byte sequences in raw_bytes and returns a
    dictionary of `start_idx -> (color, start_idx, end_idx)` for runs of
    identical bytes."""
    identical_runs = {}

    previous_color, run_start_idx = raw_bytes[0:3], 0
    for i in range(3, len(raw_bytes), 3):
        color = raw_bytes[i : i + 3]
        # Just keep going if it's the same as the previous color.
        if color == previous_color:
            continue
        # Indiciates that color did not repeat, since we only made it through
        # one loop iteration.
        if i == (run_start_idx + 3):
            run_start_idx = i
            previous_color = color
            continue
        # Color stopped repeating, put it in the runs.
        identical_runs[run_start_idx] = (previous_color, run_start_idx, i)
        run_start_idx = i
        previous_color = color

    # Make sure after the loop there's not an identical run at the end.
    if i != run_start_idx:
        identical_runs[run_start_idx] = (previous_color, run_start_idx, i + 3)

    return identical_runs


def write_reia_block(block, previous_block) -> bytearray:
    # If this block is exactly identical to the previous, we can skip encoding
    # it.
    if previous_block is not None:
        diff = ImageChops.difference(block, previous_block)
        if diff.getbbox() is None:
            return bytearray(b"\x00")

    output = bytearray(b"\x01")
    # If there is a previous block we need to compute a diff of the pixel values
    # between this and the last one.
    if previous_block is not None:
        block = ImageChops.subtract_modulo(block, previous_block)

    # Iterate over each the bytes and start calculating and writing runs.
    raw_bytes = block.tobytes("raw", "BGR")
    assert len(raw_bytes) == 32 * 32 * 3

    # First run over all the values and detect runs of the same color.
    identical_runs = find_identical_runs(raw_bytes)

    # Just so sanity check.
    pixels_encoded = 0
    # Now that we have the list of runs, we can iterate through the indices and
    # perform RLE encoding.
    i = 0
    unique_colors = []
    while i < len(raw_bytes):
        color = raw_bytes[i : i + 3]
        if i not in identical_runs:
            # Not an identical run, let's add this to the unqiue colors list and
            # we'll put them all in after we get to an identical run.
            unique_colors.append(color)
            i += 3
            continue

        # If we've got a list of unique colors built up, put them in first.
        if len(unique_colors) > 0:
            pixels_encoded += len(unique_colors)
            emit_non_repeated_colors(unique_colors, output)
            unique_colors = []

        identical_color, run_start_idx, run_end_idx = identical_runs[i]
        assert color == identical_color

        num_repeated = (run_end_idx - run_start_idx) // 3
        pixels_encoded += num_repeated
        # A negative RLE byte n indiciates the next color should be repeated
        # (-n + 1) number of times. Since min(int8) = -128 we can only repeat
        # a color a maximum of 129 times with this scheme.
        assert num_repeated <= 129

        rle_byte = -(num_repeated - 1)
        rle_byte = rle_byte.to_bytes(1, byteorder="little", signed=True)

        output.extend(rle_byte)
        output.extend(color)

        # Skip to the end index.
        i = run_end_idx
        continue
    # If we've still got unique colors going from the end of the loop, emit
    # them now.
    if len(unique_colors) > 0:
        pixels_encoded += len(unique_colors)
        emit_non_repeated_colors(unique_colors, output)

    assert pixels_encoded == 32 * 32
    return output


def emit_non_repeated_colors(unique_colors, output: bytearray):
    # A positive RLE byte n indicates (n + 1) unique values are coming.
    #
    # Since the n is a signed 8-bit int we can't be smaller than
    #   max(int8) = 127
    # so we can only encode 128 repeats at maximum.
    #
    # This loop splits up unique_colors into chunks of 129 if needed and spits
    # them out.
    for i in range(0, len(unique_colors), 128):
        colors_chunk = unique_colors[i : i + 128]

        assert len(colors_chunk) <= 128

        rle_byte = len(colors_chunk) - 1
        rle_byte = rle_byte.to_bytes(1, byteorder="little", signed=True)
        output.extend(rle_byte)
        output.extend(b"".join(colors_chunk))
